local component = require('component')

local database = component.getPrimary("database")
local me_interface = component.getPrimary("me_interface")

local D = loadfile('me_processor_lib/debug.lua')()

function handle_generic_recipe(machine, processor, recipe, amount, status_prefix)
  -- Setup hook
  if machine.setup then
    machine.setup(machine, processor, recipe, amount, status_prefix)
  end
  
  if amount <= 0 then
    print("Amount <= 0??")
    assert(false)
  end

  -- Before we do anything, we need to get the slots in recipe and sort it.
  -- The way I now define recipe slots apparently is not a real scattered indexed array according to LUA, so the sorting gets mangled
  -- Sort it here, so machines which rely on order (like the Electronics Assembler) work too.
  -- If a machine has a different slot then slot 1 as the slot to be filled first, use a slot map to define order.

  local slot_order = {}
  local n = 0
  for slot in pairs(recipe.slots) do
    n = n + 1
    slot_order[n] = slot
  end
  table.sort(slot_order)

  local slot_count = 0
  local slot_map = {}
  local ingredient = {}
  local max_ingredient_amount = 0

  for n, slot in pairs(slot_order) do
    local item = key_to_item(recipe.slots[slot])
    local item_amount = item.amount or 1
    item.amount = nil

    slot_count = slot_count + 1
    slot_map[slot_count] = slot
    ingredient[slot_count] = {item_key = item_to_key(item), amount = item_amount}
    
    if item_amount > max_ingredient_amount then
      max_ingredient_amount = item_amount
    end
  end

  -- FIXME: max_ingredient_amount ignored right now.

  if slot_count == 0 then
    print("Recipe has no ingredients?")
    assert(false)
  end

  local reserved_slots = processor.transposer:request_slots(slot_count)

  -- We could still try and do items sequentially and ask for 1 slot, but for now, just add interfaces.
  if not reserved_slots then
    print("Could not reserve slots.")
    return false
  end

  for n = 1, slot_count do
    local item = key_to_item(ingredient[n].item_key)
    item = me_interface.getItemsInNetwork({name = item.name, damage = item.damage})

    if item[1] then
      item = item[1]
    else
      print("Could not find " .. ingredient[n].item_key .. " on the network?")
      return false
    end

    local max_size = item.maxSize

    processor.transposer:request_item(ingredient[n].item_key, reserved_slots[n], (max_size > amount and amount or max_size))
  end

  -- FIXME: Move to status handler. Set telemetry in machine object.
  print("Start: " .. status_prefix .. ": " .. math.floor(amount) .. ".")

  local moved = 0

  -- FIXME: If stack_time is 2 or below, we need to go in overdrive mode.
  --        Skip the wait, just keep moving stacks in.
  --        Detect by moved if the move succeeded or failed.
  --        Do keep track of items in the system once every second, to make sure we don't run out.
  --        Also a sleep(0) is probably still needed in this mode to yield to other threads in overdrive mode.
  --        Test to see if we can overdrive a whole battery of transposers all in the same tick?
  -- FIXME: Completely disregards different item counts on the recipe. Quick & dirty to get it working.
  repeat
    local max_size = 64  -- FIXME: Obviously not.

    -- FIXME: Much wrong... Needs to be per slot or something, serious clusterfuck here with this.
    local current_amount = amount
    
    if current_amount > max_size then
      current_amount = max_size
    end

    for n = 1, slot_count do
      if machine.stack_time > 2 and not processor.transposer:wait_for_request(reserved_slots[n], current_amount, 1) then
        print("Timeout while waiting for " .. current_amount .. " of " .. database.get(reserved_slots[n].database_slot) .. ".")
        return moved
      else
        processor.transposer:move(reserved_slots[n].side, reserved_slots[n].slot, processor.side, machine.slot_map[slot_map[n]], current_amount)
      end
    end

    -- Craft hook.
    if machine.craft then
      if not machine.craft(machine, processor, recipe, current_amount, status_prefix, slot_count, reserved_slots, slot_map) then
        return
      end
      -- Craft hook is supposed to sleep, so we just yield here.
      os.sleep(0)
    else
      -- Stack time is in Ticks.
      -- FIXME: BTW. We can also keep stack_time per processor instead of per machine group.
      --        Then, if stack_time is dynamic, you can have the same machines with different speeds (due to upgrades, capacitors, etc).
      --        stack_time will not work for all machines anyway though, some recipes have different times.
      local sleep_time = 0
      if machine.stack_time > 2 then
        sleep_time = (machine.stack_time / 20 ) * (current_amount / 64)
      end

      os.sleep(sleep_time)
    end

    -- FIXME: Nonononono..... Of course not. Come on now? What ARE you thinking??
    moved = moved + current_amount

  until moved >= amount

  -- FIXME: Move to status handler. Set telemetry in machine object.
  if moved > 512 then
    print("Start: " .. status_prefix .. ": " .. math.floor(moved) .. "/" .. math.floor(amount) .. ".")
  end

  for n = 1, slot_count do
    processor.transposer:cancel_request(reserved_slots[n])
  end

  processor.transposer:free_slots(reserved_slots)
end

function processor_thread.generic(machine, processor_index)
  repeat
    request = machine.queue[1]

    if request.recipe then
      handle_generic_recipe(machine, machine.processors[processor_index], request.recipe, request.amount, request.status_prefix)
    else
      print("Queued request invalid. Ignored request.")
    end

    table.remove(machine.queue, 1)
  until #machine.queue == 0
end
